import json
from collections import OrderedDict
from datetime import datetime
from enum import Enum
from typing import Dict, Iterator, List, Optional

from botocore.exceptions import ClientError

from .boto_proxy import BotoClientProxy
from .exceptions import StackNotFound, StackNotUpdated
from .route53 import Route53


class ResourceType(str, Enum):
    route53_record_set = 'AWS::Route53::RecordSet'


class CloudFormationStack:

    """
    A single AWS Cloudformation Stack.

    See:
    http://boto3.readthedocs.io/en/latest/reference/services/cloudformation.html
    """

    def __init__(self,
                 stack_id: str,
                 name: str,
                 description: Optional[str],
                 parameters: Dict[str, str],
                 creation_time: datetime,
                 last_updated_time: Optional[datetime],
                 status: str,
                 stack_status_reason: Optional[str],
                 disable_rollback: bool,
                 notification_arns: List[str],
                 timeout_in_minutes: Optional[int],
                 capabilities: Optional[List[str]],
                 outputs: Optional[List[Dict]],
                 tags: Dict[str, str],
                 *,
                 region: Optional[str]):
        self.stack_id = stack_id
        self.name = name
        self.description = description
        self.parameters = parameters
        self.creation_time = creation_time
        self.last_updated_time = last_updated_time
        self.status = status
        self.stack_status_reason = stack_status_reason
        self.disable_rollback = disable_rollback
        self.notification_arns = notification_arns
        self.timeout_in_minutes = timeout_in_minutes
        self.capabilities = capabilities
        self.outputs = outputs
        self.tags = tags

        self.region = region

        self.__template = None

    def __repr__(self):
        return "<CloudFormationStack: {name}>".format_map(vars(self))

    @classmethod
    def from_boto_dict(cls,
                       stack: Dict,
                       region: Optional[str]=None) -> "CloudFormationStack":
        stack_id = stack['StackId']
        name = stack['StackName']
        description = stack.get('Description')
        parameters = OrderedDict([(p['ParameterKey'], p['ParameterValue'])
                                  for p in stack.get('Parameters', [])
                                  if not p.get('UsePreviousValue')])
        creation_time = stack['CreationTime']
        last_updated_time = stack.get('LastUpdatedTime')
        status = stack['StackStatus']
        stack_status_reason = stack.get('StackStatusReason')
        disable_rollback = stack['DisableRollback']
        notification_arns = stack['NotificationARNs']
        timeout_in_minutes = stack.get('TimeoutInMinutes')
        capabilities = stack.get('Capabilities')
        outputs = stack.get('Outputs')
        tags = OrderedDict([(t['Key'], t['Value']) for t in stack['Tags']])

        return cls(stack_id, name, description, parameters,
                   creation_time, last_updated_time, status,
                   stack_status_reason, disable_rollback, notification_arns,
                   timeout_in_minutes, capabilities, outputs, tags,
                   region=region)

    @classmethod
    def get_by_stack_name(cls,
                          name: str,
                          region: Optional[str]=None) -> "CloudFormationStack":
        """
        See:
        http://boto3.readthedocs.io/en/latest/reference/services/cloudformation.html#CloudFormation.Client.describe_stacks
        """
        client = BotoClientProxy('cloudformation', region)

        try:
            stacks = client.describe_stacks(StackName=name)
        except ClientError as err:
            response = err.response
            error_info = response['Error']
            error_message = error_info['Message']
            # This is not very resilient way to do this but boto API doesn't
            # provide a better way.
            if error_message == 'Stack with id {name} does not exist'.format(name=name):
                raise StackNotFound(name)
            else:
                raise
        stack = stacks['Stacks'][0]  # type: dict

        return cls.from_boto_dict(stack, region)

    @property
    def resources(self) -> Iterator:
        client = BotoClientProxy('cloudformation', self.region)
        response = client.list_stack_resources(StackName=self.stack_id)
        resources = response['StackResourceSummaries']  # type: List[Dict]
        for resource in resources:
            resource_type = resource["ResourceType"]
            if resource_type == ResourceType.route53_record_set:
                physical_resource_id = resource.get('PhysicalResourceId')
                if physical_resource_id is None:
                    # if there is no Physical Resource Id we can't fetch the
                    # record
                    continue
                records = Route53.get_records(name=resource['PhysicalResourceId'])
                for record in records:
                    if (record.set_identifier is None or
                            record.set_identifier == self.name):
                        yield record
            else:  # pragma: no cover
                # TODO implement the other resource types
                # Ignore resources that are still not implemented in manaus
                pass

    @property
    def template(self) -> Dict:
        if self.__template is None:
            client = BotoClientProxy('cloudformation', self.region)
            response = client.get_template(StackName=self.name)
            self.__template = response['TemplateBody']
        return self.__template

    def reset(self):
        self.__template = None

    def update(self):
        """
        Sends the current template to CloudFormation to update the stack
        """
        client = BotoClientProxy('cloudformation', self.region)
        parameters = [{'ParameterKey': key, 'ParameterValue': value}
                      for key, value in self.parameters.items()]
        try:
            client.update_stack(StackName=self.name,
                                TemplateBody=json.dumps(self.template),
                                Parameters=parameters,
                                Capabilities=self.capabilities)
        except ClientError as err:
            response = err.response
            error_info = response['Error']
            error_message = error_info['Message']
            if error_message == 'No updates are to be performed.':
                raise StackNotUpdated(self.name)
            else:
                raise

    def delete(self):
        client = BotoClientProxy('cloudformation', self.region)
        client.delete_stack(StackName=self.stack_id)


class CloudFormation:

    def __init__(self, region: Optional[str] = None):
        self.region = region

    def get_stacks(self, all: bool=False):
        client = BotoClientProxy('cloudformation', self.region)
        if all:
            status_filter = []
        else:
            # status_filter = [st for st in cf.valid_states if st != 'DELETE_COMPLETE']
            status_filter = [
                "CREATE_IN_PROGRESS",
                "CREATE_FAILED",
                "CREATE_COMPLETE",
                "ROLLBACK_IN_PROGRESS",
                "ROLLBACK_FAILED",
                "ROLLBACK_COMPLETE",
                "DELETE_IN_PROGRESS",
                "DELETE_FAILED",
                "UPDATE_IN_PROGRESS",
                "UPDATE_COMPLETE_CLEANUP_IN_PROGRESS",
                "UPDATE_COMPLETE",
                "UPDATE_ROLLBACK_IN_PROGRESS",
                "UPDATE_ROLLBACK_FAILED",
                "UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS",
                "UPDATE_ROLLBACK_COMPLETE"
            ]

        kwargs = {'StackStatusFilter': status_filter}
        while 'NextToken' not in kwargs or kwargs['NextToken']:
            results = client.list_stacks(**kwargs)
            for stack in results['StackSummaries']:
                stack_id = stack['StackId']
                yield CloudFormationStack.get_by_stack_name(stack_id,
                                                            region=self.region)
            kwargs['NextToken'] = results.get('NextToken')



import click
from clickclick import fatal_error
from senza.aws import resolve_security_groups

from ..cli import AccountArguments, TemplateArguments
from ..manaus import ClientError
from ..manaus.acm import ACM, ACMCertificate
from ..manaus.iam import IAM, IAMServerCertificate
from ..manaus.route53 import convert_cname_records_to_alias

SENZA_PROPERTIES = frozenset(['Domains', 'HealthCheckPath', 'HealthCheckPort', 'HealthCheckProtocol',
                              'HTTPPort', 'Name', 'SecurityGroups', 'SSLCertificateId', 'Type'])
ALLOWED_HEALTH_CHECK_PROTOCOLS = frozenset(["HTTP", "TCP", "UDP", "SSL"])
ALLOWED_LOADBALANCER_SCHEMES = frozenset(["internet-facing", "internal"])


def get_load_balancer_name(stack_name: str, stack_version: str):
    """
    Returns the name of the load balancer for the stack name and version,
    truncating the name if necessary.
    """
    # Loadbalancer name cannot exceed 32 characters, try to shorten
    l = 32 - len(stack_version) - 1
    return '{}-{}'.format(stack_name[:l], stack_version)


def get_ssl_cert(subdomain, main_zone, configuration, account_info: AccountArguments):
    ssl_cert = configuration.get('SSLCertificateId')

    if ACMCertificate.arn_is_acm_certificate(ssl_cert):
        # check if certificate really exists
        try:
            ACMCertificate.get_by_arn(account_info.Region, ssl_cert)
        except ClientError as e:
            error_msg = e.response['Error']['Message']
            fatal_error(error_msg)
    elif IAMServerCertificate.arn_is_server_certificate(ssl_cert):
        # TODO check if certificate exists
        pass
    elif ssl_cert is not None:
        certificate = IAMServerCertificate.get_by_name(account_info.Region,
                                                       ssl_cert)
        ssl_cert = certificate.arn
    elif main_zone is not None:
        if main_zone:
            iam_pattern = main_zone.lower().rstrip('.').replace('.', '-')
            name = '{sub}.{zone}'.format(sub=subdomain,
                                         zone=main_zone.rstrip('.'))
            acm = ACM(account_info.Region)
            acm_certificates = sorted(acm.get_certificates(domain_name=name),
                                      reverse=True)
        else:
            iam_pattern = ''
            acm_certificates = []
        iam = IAM(account_info.Region)
        iam_certificates = sorted(iam.get_certificates(name=iam_pattern))
        if not iam_certificates:
            # if there are no iam certificates matching the pattern
            # try to use any certificate
            iam_certificates = sorted(iam.get_certificates(), reverse=True)

        # the priority is acm_certificate first and iam_certificate second
        certificates = (acm_certificates +
                        iam_certificates)  # type: List[Union[ACMCertificate, IAMServerCertificate]]
        try:
            certificate = certificates[0]
            ssl_cert = certificate.arn
        except IndexError:
            if main_zone:
                fatal_error('Could not find any matching '
                            'SSL certificate for "{}"'.format(name))
            else:
                fatal_error('Could not find any SSL certificate')

    return ssl_cert


def get_listeners(subdomain, main_zone, configuration,
                  account_info: AccountArguments):
    ssl_cert = get_ssl_cert(subdomain, main_zone, configuration, account_info)
    return [
        {
            "PolicyNames": [],
            "SSLCertificateId": ssl_cert,
            "Protocol": "HTTPS",
            "InstancePort": configuration["HTTPPort"],
            "LoadBalancerPort": 443
        }
    ]


def component_elastic_load_balancer(definition,
                                    configuration: dict,
                                    args: TemplateArguments,
                                    info: dict,
                                    force,
                                    account_info: AccountArguments):
    lb_name = configuration["Name"]
    # domains pointing to the load balancer
    subdomain = ''
    main_zone = None
    for name, domain in configuration.get('Domains', {}).items():
        name = '{}{}'.format(lb_name, name)

        domain_name = "{0}.{1}".format(domain["Subdomain"], domain["Zone"])

        convert_cname_records_to_alias(domain_name)

        properties = {"Type": "A",
                      "Name": domain_name,
                      "HostedZoneName": domain["Zone"],
                      "AliasTarget": {"HostedZoneId": {"Fn::GetAtt": [lb_name,
                                                                      "CanonicalHostedZoneNameID"]},
                                      "DNSName": {"Fn::GetAtt": [lb_name, "DNSName"]}}}
        definition["Resources"][name] = {"Type": "AWS::Route53::RecordSet",
                                         "Properties": properties}

        if domain["Type"] == "weighted":
            definition["Resources"][name]["Properties"]['Weight'] = 0
            definition["Resources"][name]["Properties"]['SetIdentifier'] = "{0}-{1}".format(info["StackName"],
                                                                                            info["StackVersion"])
            subdomain = domain['Subdomain']
            main_zone = domain['Zone']  # type: str

    listeners = configuration.get('Listeners') or get_listeners(subdomain, main_zone, configuration, account_info)

    health_check_protocol = configuration.get('HealthCheckProtocol') or 'HTTP'

    if health_check_protocol not in ALLOWED_HEALTH_CHECK_PROTOCOLS:
        raise click.UsageError('Protocol "{}" is not supported for LoadBalancer'.format(health_check_protocol))

    health_check_path = configuration.get("HealthCheckPath") or '/health'
    health_check_port = configuration.get("HealthCheckPort") or configuration["HTTPPort"]

    health_check_target = "{0}:{1}{2}".format(health_check_protocol,
                                              health_check_port,
                                              health_check_path)

    if configuration.get('NameSuffix'):
        version = '{}-{}'.format(info["StackVersion"],
                                 configuration['NameSuffix'])
        loadbalancer_name = get_load_balancer_name(info["StackName"], version)
        del(configuration['NameSuffix'])
    else:
        loadbalancer_name = get_load_balancer_name(info["StackName"],
                                                   info["StackVersion"])

    loadbalancer_scheme = configuration.get('Scheme') or 'internal'

    if loadbalancer_scheme == 'internet-facing':
        click.secho('You are deploying an internet-facing ELB that will be '
                    'publicly accessible! You should have OAUTH2 and HTTPS '
                    'in place!', bold=True, err=True)

    if loadbalancer_scheme not in ALLOWED_LOADBALANCER_SCHEMES:
        raise click.UsageError('Scheme "{}" is not supported for LoadBalancer'.format(loadbalancer_scheme))

    if loadbalancer_scheme == "internal":
        loadbalancer_subnet_map = "LoadBalancerInternalSubnets"
    else:
        loadbalancer_subnet_map = "LoadBalancerSubnets"

    # load balancer
    definition["Resources"][lb_name] = {
        "Type": "AWS::ElasticLoadBalancing::LoadBalancer",
        "Properties": {
            "Scheme": loadbalancer_scheme,
            "Subnets": {"Fn::FindInMap": [loadbalancer_subnet_map, {"Ref": "AWS::Region"}, "Subnets"]},
            "HealthCheck": {
                "HealthyThreshold": "2",
                "UnhealthyThreshold": "2",
                "Interval": "10",
                "Timeout": "5",
                "Target": health_check_target
            },
            "Listeners": listeners,
            "ConnectionDrainingPolicy": {
                "Enabled": True,
                "Timeout": 60
            },
            "CrossZone": "true",
            "LoadBalancerName": loadbalancer_name,
            "SecurityGroups": resolve_security_groups(configuration["SecurityGroups"], args.region),
            "Tags": [
                # Tag "Name"
                {
                    "Key": "Name",
                    "Value": "{0}-{1}".format(info["StackName"], info["StackVersion"])
                },
                # Tag "StackName"
                {
                    "Key": "StackName",
                    "Value": info["StackName"],
                },
                # Tag "StackVersion"
                {
                    "Key": "StackVersion",
                    "Value": info["StackVersion"]
                }
            ]
        }
    }
    for key, val in configuration.items():
        # overwrite any specified properties, but
        # ignore our special Senza properties as they are not supported by CF
        if key not in SENZA_PROPERTIES:
            definition['Resources'][lb_name]['Properties'][key] = val

    return definition



"""
Senza's root command with tasks and flags common to all sub-commands
"""

import sys
import time
from distutils.version import LooseVersion
from pathlib import Path
from typing import Optional

import click
import requests
import senza
from clickclick import AliasedGroup, warning

from ..arguments import GLOBAL_OPTIONS, region_option
from ..error_handling import sentry

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

PYPI_URL = "https://pypi.python.org/pypi/stups-senza/json"
ONE_DAY = 86400  # seconds


def get_latest_version_from_disk() -> Optional[LooseVersion]:
    """
    Tries to read a cached latest version from the disk returning None if the
    file doesn't exist or if it's older than 24 hours
    """
    version_cache = Path(click.get_app_dir('senza')) / 'pypi_version'
    now = time.time()
    latest_version = None
    if (version_cache.exists() and
            now - version_cache.stat().st_mtime < ONE_DAY):
        with version_cache.open() as version_cache_file:
            str_version = version_cache_file.read()
            if str_version:
                latest_version = LooseVersion(str_version)
    return latest_version


def get_latest_version_from_pypi() -> Optional[LooseVersion]:
    """
    Gets the latest release version pypi api using distutils order
    to sort the releases (same as pip).
    """
    try:
        pypi_response = requests.get(PYPI_URL, timeout=1)
    except requests.Timeout:
        return None

    # the potential exception is not caught here but it's caught in
    # check_senza_version and pushed to sentry if it is configured
    pypi_response.raise_for_status()
    pypi_data = pypi_response.json()
    releases = pypi_data['releases']
    versions = [LooseVersion(version) for version in releases.keys()]
    return sorted(versions)[-1]


def get_latest_version() -> Optional[LooseVersion]:
    """
    Gets the latest version either from the file cache or from pip.

    If the file cache exists it will be valid for 24 hours.
    """
    version_cache = Path(click.get_app_dir('senza')) / 'pypi_version'
    latest_version = (get_latest_version_from_disk() or
                      get_latest_version_from_pypi())

    if latest_version is not None:
        try:
            version_cache.parent.mkdir(parents=True)
        except FileExistsError:
            # this try...except can be replaced with exist_ok=True when
            # we drop python3.4 support
            pass

        with version_cache.open('w') as version_cache_file:
            version_cache_file.write(str(latest_version))
    return latest_version


def check_senza_version(current_version: str):
    """
    Checks if senza is updated and prints a warning with instructions to update
    if it's not.
    """
    if not sys.stdout.isatty():
        return
    current_version = LooseVersion(current_version)
    try:
        latest_version = get_latest_version()
    except Exception:
        if sentry is not None:
            sentry.captureException()
        return

    if latest_version is not None and current_version < latest_version:
        if __file__.startswith('/home'):
            # if it's installed in the user folder
            cmd = "pip install --upgrade stups-senza"
        else:
            cmd = "sudo pip install --upgrade stups-senza"
        warning("Your senza version ({current}) is outdated. "
                "Please install the new one using '{cmd}'".format(current=current_version,
                                                                  cmd=cmd))


def print_version(ctx, param, value):
    """
    Prints current senza version and checks if it's the latest one.
    """
    assert param.name == "version"
    if not value or ctx.resilient_parsing:
        return

    click.echo('Senza {}'.format(senza.__version__))
    # this needs to be here since when this is called cli() is not
    check_senza_version(senza.__version__)
    ctx.exit()


@click.group(cls=AliasedGroup, context_settings=CONTEXT_SETTINGS)
@click.option('-V', '--version',
              is_flag=True, callback=print_version, expose_value=False,
              is_eager=True,
              help='Print the current version number and exit.')
@region_option
def cli(region):
    """
    Senza's root command.

    It checks the version and sets the region global option before executing
    the sub-commands.

    Sub command can be added by using `cli.add_command(SUB_COMMAND_FUNCTION)`
    or using the `@cli.command()` decorator
    """
    check_senza_version(senza.__version__)
    GLOBAL_OPTIONS['region'] = region



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import inspect

import setuptools
from setuptools.command.test import test as TestCommand
from setuptools import setup

if sys.version_info < (3, 4, 0):
    sys.stderr.write('FATAL: STUPS Senza needs to be run with Python 3.4+\n')
    sys.exit(1)

__location__ = os.path.join(os.getcwd(), os.path.dirname(inspect.getfile(inspect.currentframe())))


def read_version(package):
    with open(os.path.join(package, '__init__.py'), 'r') as fd:
        for line in fd:
            if line.startswith('__version__ = '):
                return line.split()[-1].strip().strip("'")

NAME = 'stups-senza'
MAIN_PACKAGE = 'senza'
VERSION = read_version(MAIN_PACKAGE)
DESCRIPTION = 'AWS Cloud Formation deployment CLI'
LICENSE = 'Apache License 2.0'
URL = 'https://github.com/zalando-stups/senza'
AUTHOR = 'Henning Jacobs'
EMAIL = 'henning.jacobs@zalando.de'
KEYWORDS = 'aws cloud formation cf elb ec2 stups immutable stacks route53 boto'

COVERAGE_XML = True
COVERAGE_HTML = False
JUNIT_XML = True

# Add here all kinds of additional classifiers as defined under
# https://pypi.python.org/pypi?%3Aaction=list_classifiers
CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: Implementation :: CPython',
]

CONSOLE_SCRIPTS = ['senza = senza.cli:main']


class PyTest(TestCommand):

    user_options = [('cov=', None, 'Run coverage'), ('cov-xml=', None, 'Generate junit xml report'), ('cov-html=',
                    None, 'Generate junit html report'), ('junitxml=', None, 'Generate xml of test results')]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.cov = None
        self.cov_xml = False
        self.cov_html = False
        self.junitxml = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        if self.cov is not None:
            self.cov = ['--cov', self.cov, '--cov-report', 'term-missing']
            if self.cov_xml:
                self.cov.extend(['--cov-report', 'xml'])
            if self.cov_html:
                self.cov.extend(['--cov-report', 'html'])
        if self.junitxml is not None:
            self.junitxml = ['--junitxml', self.junitxml]

    def run_tests(self):
        try:
            import pytest
        except:
            raise RuntimeError('py.test is not installed, run: pip install pytest')
        params = {'args': self.test_args}
        if self.cov:
            params['args'] += self.cov
        if self.junitxml:
            params['args'] += self.junitxml
        errno = pytest.main(**params)
        sys.exit(errno)


def get_install_requirements(path):
    content = open(os.path.join(__location__, path)).read()
    return [req for req in content.split('\\n') if req != '']


def read(fname):
    return open(os.path.join(__location__, fname)).read()


def setup_package():
    # Assemble additional setup commands
    cmdclass = {}
    cmdclass['test'] = PyTest

    install_reqs = get_install_requirements('requirements.txt')

    command_options = {'test': {'test_suite': ('setup.py', 'tests'), 'cov': ('setup.py', MAIN_PACKAGE)}}
    if JUNIT_XML:
        command_options['test']['junitxml'] = 'setup.py', 'junit.xml'
    if COVERAGE_XML:
        command_options['test']['cov_xml'] = 'setup.py', True
    if COVERAGE_HTML:
        command_options['test']['cov_html'] = 'setup.py', True

    setup(
        name=NAME,
        version=VERSION,
        url=URL,
        description=DESCRIPTION,
        author=AUTHOR,
        author_email=EMAIL,
        license=LICENSE,
        keywords=KEYWORDS,
        long_description=read('README.rst'),
        classifiers=CLASSIFIERS,
        test_suite='tests',
        packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
        install_requires=install_reqs,
        setup_requires=['flake8'],
        cmdclass=cmdclass,
        tests_require=['pytest-cov', 'pytest', 'mock'],
        command_options=command_options,
        entry_points={'console_scripts': CONSOLE_SCRIPTS,
                      'senza.templates': ['bgapp = senza.templates.bgapp',
                                          'postgresapp = senza.templates.postgresapp',
                                          'rediscluster = senza.templates.rediscluster',
                                          'redisnode = senza.templates.redisnode',
                                          'webapp = senza.templates.webapp']},
    )


if __name__ == '__main__':
    setup_package()



import click
import re
from senza.aws import resolve_security_groups, resolve_topic_arn, resolve_referenced_resource
from senza.utils import ensure_keys
from senza.components.iam_role import get_merged_policies

# properties evaluated by Senza
SENZA_PROPERTIES = frozenset(['SecurityGroups', 'Tags'])

# additional CF properties which can be overwritten
ADDITIONAL_PROPERTIES = {
    'AWS::AutoScaling::LaunchConfiguration': frozenset(['BlockDeviceMappings', 'IamInstanceProfile', 'SpotPrice']),
    'AWS::AutoScaling::AutoScalingGroup': frozenset(['MetricsCollection', 'TargetGroupARNs',
                                                     'TerminationPolicies', 'PlacementGroup'])
}


def component_auto_scaling_group(definition, configuration, args, info, force, account_info):
    definition = ensure_keys(definition, "Resources")

    # launch configuration
    config_name = configuration["Name"] + "Config"
    definition["Resources"][config_name] = {
        "Type": "AWS::AutoScaling::LaunchConfiguration",
        "Properties": {
            "InstanceType": configuration["InstanceType"],
            "ImageId": {"Fn::FindInMap": ["Images", {"Ref": "AWS::Region"}, configuration["Image"]]},
            "AssociatePublicIpAddress": configuration.get('AssociatePublicIpAddress', False),
            "EbsOptimized": configuration.get('EbsOptimized', False)
        }
    }

    if 'IamRoles' in configuration:
        logical_id = configuration['Name'] + 'InstanceProfile'
        roles = configuration['IamRoles']
        if len(roles) > 1:
            for role in roles:
                if isinstance(role, dict):
                    raise click.UsageError('Cannot merge policies of Cloud Formation references ({"Ref": ".."}): ' +
                                           'You can use at most one IAM role with "Ref".')
            logical_role_id = configuration['Name'] + 'Role'
            definition['Resources'][logical_role_id] = {
                'Type': 'AWS::IAM::Role',
                'Properties': {
                    "AssumeRolePolicyDocument": {
                        "Version": "2012-10-17",
                        "Statement": [
                            {
                                "Effect": "Allow",
                                "Principal": {
                                    "Service": ["ec2.amazonaws.com"]
                                },
                                "Action": ["sts:AssumeRole"]
                            }
                        ]
                    },
                    'Path': '/',
                    'Policies': get_merged_policies(roles)
                }
            }
            instance_profile_roles = [{'Ref': logical_role_id}]
        elif isinstance(roles[0], dict):
            instance_profile_roles = [resolve_referenced_resource(roles[0], args.region)]
        else:
            instance_profile_roles = roles
        definition['Resources'][logical_id] = {
            'Type': 'AWS::IAM::InstanceProfile',
            'Properties': {
                'Path': '/',
                'Roles': instance_profile_roles
            }
        }
        definition["Resources"][config_name]["Properties"]["IamInstanceProfile"] = {'Ref': logical_id}

    if "SecurityGroups" in configuration:
        definition["Resources"][config_name]["Properties"]["SecurityGroups"] = \
            resolve_security_groups(configuration["SecurityGroups"], args.region)

    if "UserData" in configuration:
        definition["Resources"][config_name]["Properties"]["UserData"] = {
            "Fn::Base64": configuration["UserData"]
        }

    # auto scaling group
    asg_name = configuration["Name"]
    asg_success = ["1", "PT15M"]
    if "AutoScaling" in configuration:
        if "SuccessRequires" in configuration["AutoScaling"]:
            asg_success = normalize_asg_success(configuration["AutoScaling"]["SuccessRequires"])

    tags = [
        # Tag "Name"
        {
            "Key": "Name",
            "PropagateAtLaunch": True,
            "Value": "{0}-{1}".format(info["StackName"], info["StackVersion"])
        },
        # Tag "StackName"
        {
            "Key": "StackName",
            "PropagateAtLaunch": True,
            "Value": info["StackName"],
        },
        # Tag "StackVersion"
        {
            "Key": "StackVersion",
            "PropagateAtLaunch": True,
            "Value": info["StackVersion"]
        }
    ]

    if "Tags" in configuration:
        for tag in configuration["Tags"]:
            tags.append({
                "Key": tag["Key"],
                "PropagateAtLaunch": True,
                "Value": tag["Value"]
            })

    definition["Resources"][asg_name] = {
        "Type": "AWS::AutoScaling::AutoScalingGroup",
        # wait to get a signal from an amount of servers to signal that it booted
        "CreationPolicy": {
            "ResourceSignal": {
                "Count": asg_success[0],
                "Timeout": asg_success[1]
            }
        },
        "Properties": {
            # for our operator some notifications
            "LaunchConfigurationName": {"Ref": config_name},
            "VPCZoneIdentifier": {"Fn::FindInMap": ["ServerSubnets", {"Ref": "AWS::Region"}, "Subnets"]},
            "Tags": tags
        }
    }

    asg_properties = definition["Resources"][asg_name]["Properties"]

    if "OperatorTopicId" in info:
        asg_properties["NotificationConfiguration"] = {
            "NotificationTypes": [
                "autoscaling:EC2_INSTANCE_LAUNCH",
                "autoscaling:EC2_INSTANCE_LAUNCH_ERROR",
                "autoscaling:EC2_INSTANCE_TERMINATE",
                "autoscaling:EC2_INSTANCE_TERMINATE_ERROR"
            ],
            "TopicARN": resolve_topic_arn(args.region, info["OperatorTopicId"])
        }

    default_health_check_type = 'EC2'

    if "ElasticLoadBalancer" in configuration:
        if isinstance(configuration["ElasticLoadBalancer"], str):
            asg_properties["LoadBalancerNames"] = [{"Ref": configuration["ElasticLoadBalancer"]}]
        elif isinstance(configuration["ElasticLoadBalancer"], list):
            asg_properties["LoadBalancerNames"] = [{'Ref': ref} for ref in configuration["ElasticLoadBalancer"]]
        # use ELB health check by default
        default_health_check_type = 'ELB'
    elif "ElasticLoadBalancerV2" in configuration:
        if isinstance(configuration["ElasticLoadBalancerV2"], str):
            asg_properties["TargetGroupARNs"] = [{"Ref": configuration["ElasticLoadBalancerV2"] + 'TargetGroup'}]
        elif isinstance(configuration["ElasticLoadBalancerV2"], list):
            asg_properties["TargetGroupARNs"] = [
                {'Ref': ref} for ref in configuration["ElasticLoadBalancerV2"] + 'TargetGroup'
            ]
        # use ELB health check by default
        default_health_check_type = 'ELB'

    asg_properties['HealthCheckType'] = configuration.get('HealthCheckType', default_health_check_type)
    asg_properties['HealthCheckGracePeriod'] = configuration.get('HealthCheckGracePeriod', 300)

    if "AutoScaling" in configuration:
        as_conf = configuration["AutoScaling"]
        asg_properties["MaxSize"] = as_conf["Maximum"]
        asg_properties["MinSize"] = as_conf["Minimum"]
        asg_properties["DesiredCapacity"] = max(int(as_conf["Minimum"]), int(as_conf.get('DesiredCapacity', 1)))

        default_scaling_adjustment = as_conf.get("ScalingAdjustment", 1)
        default_cooldown = as_conf.get("Cooldown", "60")

        # ScaleUp policy
        scaling_up_adjustment = int(as_conf.get("ScaleUpAdjustment", default_scaling_adjustment))
        definition["Resources"][asg_name + "ScaleUp"] = {
            "Type": "AWS::AutoScaling::ScalingPolicy",
            "Properties": {
                "AdjustmentType": "ChangeInCapacity",
                "ScalingAdjustment": str(scaling_up_adjustment),
                "Cooldown": str(as_conf.get("ScaleUpCooldown", default_cooldown)),
                "AutoScalingGroupName": {
                    "Ref": asg_name
                }
            }
        }

        # ScaleDown policy
        scaling_down_adjustment = int(as_conf.get("ScaleDownAdjustment", default_scaling_adjustment))
        definition["Resources"][asg_name + "ScaleDown"] = {
            "Type": "AWS::AutoScaling::ScalingPolicy",
            "Properties": {
                "AdjustmentType": "ChangeInCapacity",
                "ScalingAdjustment": str((-1) * scaling_down_adjustment),
                "Cooldown": str(as_conf.get("ScaleDownCooldown", default_cooldown)),
                "AutoScalingGroupName": {
                    "Ref": asg_name
                }
            }
        }

        if "MetricType" in as_conf:
            metric_type = as_conf["MetricType"]
            metricfns = {
                "CPU": metric_cpu,
                "NetworkIn": metric_network,
                "NetworkOut": metric_network
            }
            # lowercase cpu is an acceptable metric, be compatible
            if metric_type.lower() not in map(lambda t: t.lower(), metricfns.keys()):
                raise click.UsageError('Auto scaling MetricType "{}" not supported.'.format(metric_type))
            metricfn = metricfns[metric_type]
            definition = metricfn(asg_name, definition, as_conf, args, info, force)
    else:
        asg_properties["MaxSize"] = 1
        asg_properties["MinSize"] = 1

    for res in (config_name, asg_name):
        props = definition['Resources'][res]['Properties']
        additional_cf_properties = ADDITIONAL_PROPERTIES.get(definition['Resources'][res]['Type'])
        properties_allowed_to_overwrite = (set(props.keys()) - SENZA_PROPERTIES) | additional_cf_properties
        for key in properties_allowed_to_overwrite:
            if key in configuration:
                props[key] = configuration[key]

    return definition

duration_regex = r'^(?:\d+[hH])?(?:\d+[mM])?(?:\d+[sS])?$'
duration_split_regex = r'(\d+[hHmMsS])'


def to_iso8601_duration(duration):
    if duration and re.match(duration_regex, duration):
        durations = [d.upper() for d in re.split(duration_split_regex, duration)]
        return "PT" + "".join(durations)
    else:
        raise click.UsageError("Unknown duration {}. Use something like 15m.".format(duration))


def normalize_asg_success(success):
    count = "1"
    duration = "PT15M"

    # if it's falsy, return defaults
    if not success:
        return [count, duration]

    # if it's int, use as instance count
    if isinstance(success, int):
        return [str(success), duration]

    try:
        # try to parse as int
        count = int(success)
        # if it works, use as instance count
        return [success, duration]
    except ValueError:
        # ok did not work, try to parse
        if "within" in success:
            instance, time = success.split("within")
            return [instance.strip(), to_iso8601_duration(time.strip())]
        else:
            msg = 'Unknown ASG success requirement "{}". Use something like "1 within 10m".'
            raise click.UsageError(msg.format(success))


def normalize_network_threshold(threshold):
    unit = "Bytes"
    if threshold is None:
        return []
    if isinstance(threshold, int):
        return [str(threshold), unit]
    amount = 1024
    shortcuts = {
        "B": "Bytes",
        "KB": "Kilobytes",
        "MB": "Megabytes",
        "GB": "Gigabytes",
        "TB": "Terabytes"
    }
    try:
        # if someone write just Threshold: 10
        amount = int(threshold)
        return [threshold, unit]
    except ValueError:
        # check if there is a space as if somebody wrote Threshold: 20 GB
        if " " in threshold:
            # okay, so split it
            amount, unit = threshold.split()
            if unit in shortcuts:
                unit = shortcuts[unit]
            allowed_units = shortcuts.values()
            if unit not in allowed_units:
                raise click.UsageError("Network threshold unit must be one of {}".format(list(allowed_units)))
        else:
            raise click.UsageError('Unknown network threshold "{}". Use something like "20 GB".'.format(threshold))
    return [amount, unit]


def metric_network(asg_name, definition, configuration, args, info, force):
    period = int(configuration.get("Period", 300))
    evaluation_periods = int(configuration.get("EvaluationPeriods", 2))
    statistic = configuration.get("Statistic", "Average")
    scale_up_threshold = normalize_network_threshold(configuration["ScaleUpThreshold"])
    scale_down_threshold = normalize_network_threshold(configuration["ScaleDownThreshold"])

    if "ScaleUpThreshold" in configuration:
        definition["Resources"][asg_name + "NetworkAlarmHigh"] = {
            "Type": "AWS::CloudWatch::Alarm",
            "Properties": {
                "MetricName": configuration["MetricType"],
                "Namespace": "AWS/EC2",
                "Period": str(period),
                "Threshold": scale_up_threshold[0],
                "Unit": scale_up_threshold[1],
                "EvaluationPeriods": str(evaluation_periods),
                "Statistic": statistic,
                "ComparisonOperator": "GreaterThanThreshold",
                "Dimensions": [
                    {
                        "Name": "AutoScalingGroupName",
                        "Value": {"Ref": asg_name}
                    }
                ],
                "AlarmDescription": "Scale-up if {} > {} {} for {} minutes ({})".format(
                    configuration["MetricType"],
                    scale_up_threshold[0],
                    scale_up_threshold[1],
                    (period / 60) * evaluation_periods,
                    statistic
                ),
                "AlarmActions": [
                    {"Ref": asg_name + "ScaleUp"}
                ]
            }
        }

    if "ScaleDownThreshold" in configuration:
        definition["Resources"][asg_name + "NetworkAlarmLow"] = {
            "Type": "AWS::CloudWatch::Alarm",
            "Properties": {
                "MetricName": configuration["MetricType"],
                "Namespace": "AWS/EC2",
                "Threshold": scale_down_threshold[0],
                "Unit": scale_down_threshold[1],
                "Period": str(period),
                "EvaluationPeriods": str(evaluation_periods),
                "Statistic": statistic,
                "ComparisonOperator": "LessThanThreshold",
                "Dimensions": [
                    {
                        "Name": "AutoScalingGroupName",
                        "Value": {"Ref": asg_name}
                    }
                ],
                "AlarmDescription": "Scale-down if {} < {} {} for {} minutes ({})".format(
                    configuration["MetricType"],
                    scale_down_threshold[0],
                    scale_down_threshold[1],
                    (period / 60) * evaluation_periods,
                    statistic
                ),
                "AlarmActions": [
                    {"Ref": asg_name + "ScaleDown"}
                ]
            }
        }

    return definition


def metric_cpu(asg_name, definition, configuration, args, info, force):
    period = int(configuration.get("Period", 300))
    evaluation_periods = int(configuration.get("EvaluationPeriods", 2))
    statistic = configuration.get("Statistic", "Average")
    if "ScaleUpThreshold" in configuration:
        definition["Resources"][asg_name + "CPUAlarmHigh"] = {
            "Type": "AWS::CloudWatch::Alarm",
            "Properties": {
                "MetricName": "CPUUtilization",
                "Namespace": "AWS/EC2",
                "Period": str(period),
                "EvaluationPeriods": str(evaluation_periods),
                "Statistic": statistic,
                "Threshold": configuration["ScaleUpThreshold"],
                "ComparisonOperator": "GreaterThanThreshold",
                "Dimensions": [
                    {
                        "Name": "AutoScalingGroupName",
                        "Value": {"Ref": asg_name}
                    }
                ],
                "AlarmDescription": "Scale-up if CPU > {}% for {} minutes ({})".format(
                    configuration["ScaleUpThreshold"],
                    (period / 60) * evaluation_periods,
                    statistic
                ),
                "AlarmActions": [
                    {"Ref": asg_name + "ScaleUp"}
                ]
            }
        }

    if "ScaleDownThreshold" in configuration:
        definition["Resources"][asg_name + "CPUAlarmLow"] = {
            "Type": "AWS::CloudWatch::Alarm",
            "Properties": {
                "MetricName": "CPUUtilization",
                "Namespace": "AWS/EC2",
                "Period": str(period),
                "EvaluationPeriods": str(evaluation_periods),
                "Statistic": statistic,
                "Threshold": configuration["ScaleDownThreshold"],
                "ComparisonOperator": "LessThanThreshold",
                "Dimensions": [
                    {
                        "Name": "AutoScalingGroupName",
                        "Value": {"Ref": asg_name}
                    }
                ],
                "AlarmDescription": "Scale-down if CPU < {}% for {} minutes ({})".format(
                    configuration["ScaleDownThreshold"],
                    (period / 60) * evaluation_periods,
                    statistic
                ),
                "AlarmActions": [
                    {"Ref": asg_name + "ScaleDown"}
                ]
            }
        }

    return definition



"""
ELB_ related classes and functions.

For more information see the `ELB documentation`_ and the `boto3 documentation`_

.. _ELB: https://aws.amazon.com/elasticloadbalancing/
.. _ELB documentation: https://aws.amazon.com/documentation/elastic-load-balancing/
.. _boto3 documentation: http://boto3.readthedocs.io/en/latest/reference/services/elb.html
"""

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from .boto_proxy import BotoClientProxy
from .exceptions import ELBNotFound
from .route53 import Route53HostedZone


class ELBScheme(str, Enum):
    """
    DNS record type
    """
    internet_facing = 'internet-facing'
    internal = 'internal'

    @classmethod
    def from_str(cls, value: str) -> "ELBScheme":
        """
        Returns the an attribute based on a string as returned by boto
        """
        value = value.replace('-', '_')
        return cls[value]


class ELBHealthCheck:

    """
    Represents the HealthCheck section of an ELB.

    See `HealthCheck` key in:
    http://boto3.readthedocs.io/en/latest/reference/services/elb.html#ElasticLoadBalancing.Client.describe_load_balancers
    """

    def __init__(self,
                 target: str,
                 interval: int,
                 timeout: int,
                 unhealthy_threshold: int,
                 healthy_threshold: int):
        self.target = target
        self.interval = interval
        self.timeout = timeout
        self.unhealthy_threshold = unhealthy_threshold
        self.healthy_threshold = healthy_threshold

    @classmethod
    def from_boto_dict(cls, health_check: Dict[str, Any]) -> "ELBHealthCheck":
        """
        Instantiates a `ELBHealthCheck` object from the dictionary returned
        by boto3.
        """
        target = health_check['Target']
        interval = health_check['Interval']
        timeout = health_check['Timeout']
        unhealthy_threshold = health_check['UnhealthyThreshold']
        healthy_threshold = health_check['HealthyThreshold']

        cls(target, interval, timeout, unhealthy_threshold, healthy_threshold)


class ELBListener:

    """
    Represents the ELBListener section of an ELB.

    See `ELBListener` key in:
    http://boto3.readthedocs.io/en/latest/reference/services/elb.html#ElasticLoadBalancing.Client.describe_load_balancers
    """

    def __init__(self,
                 protocol: str,
                 load_balancer_port: int,
                 instance_protocol: str,
                 instance_port: int,
                 ssl_certificate_id: str):
        self.protocol = protocol
        self.load_balancer_port = load_balancer_port
        self.instance_protocol = instance_protocol
        self.instance_port = instance_port
        self.ssl_certificate_id = ssl_certificate_id

    @classmethod
    def from_boto_dict(cls, listener) -> "ELBListener":
        """
        Instantiates a `ELBListener` object from the dictionary returned
        by boto3.
        """

        protocol = listener['Protocol']
        load_balancer_port = listener['LoadBalancerPort']
        instance_protocol = listener['InstanceProtocol']
        instance_port = listener['InstancePort']
        ssl_certificate_id = listener.get('SSLCertificateId')

        return cls(protocol, load_balancer_port, instance_protocol,
                   instance_port, ssl_certificate_id)


class ELB:

    """
    Elastic Load Balancer.

    See:
    http://boto3.readthedocs.io/en/latest/reference/services/elb.html#ElasticLoadBalancing.Client.describe_load_balancers
    """

    def __init__(self,
                 name: str,
                 dns_name: str,
                 hosted_zone_name: str,
                 hosted_zone_id: str,
                 listener_descriptions: dict,
                 policies: dict,
                 backend_server_descriptions: list,
                 availability_zones: List[str],
                 subnets: List[str],
                 instance_ids: List[str],
                 vpc_id: str,
                 health_check: ELBHealthCheck,
                 source_security_group: Dict[str, str],
                 security_groups: List[str],
                 created_time: datetime,
                 scheme: ELBScheme,
                 listeners: Optional[ELBListener]=None,
                 region: Optional[str]=None):
        self.name = name
        self.dns_name = dns_name
        self.hosted_zone_name = hosted_zone_name
        self.hosted_zone_id = hosted_zone_id
        self.policies = policies
        self.backend_server_descriptions = backend_server_descriptions
        self.availability_zones = availability_zones
        self.subnets = subnets
        self.vpc_id = vpc_id
        self.instance_ids = instance_ids
        self.health_check = health_check
        self.source_security_group = source_security_group
        self.security_groups = security_groups
        self.created_time = created_time
        self.scheme = scheme

        self.hosted_zone = Route53HostedZone(name=hosted_zone_name,
                                             id=hosted_zone_id)

        if listeners is None:
            listeners = [ELBListener.from_boto_dict(each['Listener'])
                         for each in listener_descriptions]
        self.listeners = listeners

        if region is None:
            _, region, _ = dns_name.split('.', maxsplit=2)

        self.region = region

    @classmethod
    def from_boto_dict(cls, load_balancer) -> "ELB":
        """
        Instantiates a `ELB` object from the dictionary returned
        by `boto3.client('elb').describe_load_balancers`.
        """

        name = load_balancer['LoadBalancerName']
        dns_name = load_balancer['DNSName']
        hosted_zone_name = load_balancer.get('CanonicalHostedZoneName')
        hosted_zone_id = load_balancer['CanonicalHostedZoneNameID']
        listener_descriptions = load_balancer['ListenerDescriptions']
        policies = load_balancer['Policies']
        backend_server_descriptions = load_balancer['BackendServerDescriptions']
        availability_zones = load_balancer['AvailabilityZones']
        subnets = load_balancer['Subnets']
        vpc_id = load_balancer['VPCId']
        instance_ids = load_balancer['Instances']
        health_check = ELBHealthCheck.from_boto_dict(load_balancer['HealthCheck'])
        source_security_group = load_balancer['SourceSecurityGroup']
        security_groups = load_balancer['SecurityGroups']
        created_time = load_balancer['CreatedTime']
        scheme = ELBScheme.from_str(load_balancer['Scheme'])

        return cls(name, dns_name, hosted_zone_name, hosted_zone_id,
                   listener_descriptions, policies,
                   backend_server_descriptions, availability_zones,
                   subnets, vpc_id, instance_ids, health_check,
                   source_security_group, security_groups, created_time,
                   scheme)

    @classmethod
    def get_by_dns_name(cls, dns_name: str) -> "ELB":
        """
        Gets a `ELB` by its DNS name and raises and `ELBNotFound` exception
        if such an ELB doesn't exist.

        :raises: ELBNotFound
        """
        _, region, _ = dns_name.split('.', maxsplit=2)
        client = BotoClientProxy('elb', region)

        response = client.describe_load_balancers()
        next_marker = response.get('NextMarker')
        load_balancers = response['LoadBalancerDescriptions']  # type: List
        while next_marker:
            response = client.describe_load_balancers(Marker=next_marker)
            next_marker = response.get('NextMarker')
            load_balancers.extend(response['LoadBalancerDescriptions'])

        for load_balancer in load_balancers:
            if load_balancer['DNSName'] == dns_name:
                return cls.from_boto_dict(load_balancer)

        raise ELBNotFound(dns_name)



from collections.abc import MutableMapping
from pathlib import Path
from typing import Dict, Tuple

import yaml
from click import get_app_dir

from .exceptions import InvalidConfigKey


class Configuration(MutableMapping):

    def __init__(self, path: Path):
        self.config_path = path

    def __iter__(self):
        yield from self.raw_dict

    def __len__(self):
        return len(self.raw_dict)

    def __getitem__(self, key: str) -> str:
        section, sub_key = self.__split_key(key)
        return self.raw_dict[section][sub_key]

    def __setitem__(self, key: str, value):
        section, sub_key = self.__split_key(key)
        configuration = self.raw_dict

        if section not in configuration:
            configuration[section] = {}
        configuration[section][sub_key] = str(value)
        self.__save(configuration)

    def __delitem__(self, key):
        section, sub_key = self.__split_key(key)
        cfg = self.raw_dict
        del cfg[section][sub_key]
        self.__save(cfg)

    @staticmethod
    def __split_key(key: str) -> Tuple[str, str]:
        """
        Splits the full key in section and subkey
        """
        try:
            section, sub_key = key.split('.', 1)
        except ValueError:
            # error message inspired by git config
            raise InvalidConfigKey('key does not contain '
                                   'a section: {}'.format(key))
        return section, sub_key

    def __save(self, cfg):
        """
        Saves the configuration in the configuration path, creating the
        directory if necessary.
        """
        try:
            self.config_path.parent.mkdir(parents=True)
        except FileExistsError:
            # this try...except can be replaced with exist_ok=True when
            # we drop python3.4 support
            pass
        with self.config_path.open('w+') as config_file:
            yaml.safe_dump(cfg, config_file,
                           default_flow_style=False)

    @property
    def raw_dict(self) -> Dict[str, Dict[str, str]]:
        """
        Returns a dict with the configuration data as stored in config.yaml
        """
        try:
            with self.config_path.open() as config_file:
                cfg = yaml.safe_load(config_file)
        except FileNotFoundError:
            cfg = {}
        return cfg

configuration = Configuration(Path(get_app_dir('senza')) / "config.yaml")



class ManausException(Exception):
    """
    Base class for Manaus exceptions
    """


class InvalidState(ManausException):
    """
    Exception raised when executing an action would try to change a stack
    to an invalid state
    """

    def __init__(self, reason: str):
        super().__init__('Invalid State: {}'.format(reason))


class ELBNotFound(ManausException):
    """
    Error raised when the ELB is not found
    """

    def __init__(self, domain_name: str):
        super().__init__('ELB not found: {}'.format(domain_name))


class HostedZoneNotFound(ManausException):
    """
    Error raised when the Route53 hosted zone is not found
    """

    def __init__(self, name: str):
        super().__init__('Hosted Zone not found: {}'.format(name))


class RecordNotFound(ManausException):
    """
    Error raised when the Route53 record is not found
    """

    def __init__(self, name: str):
        super().__init__('Route 53 Record not found: {}'.format(name))


class StackNotFound(ManausException):
    """
    Error raised when the CloudFormation Stack is not found
    """

    def __init__(self, name: str):
        super().__init__('CloudFormation Stack not found: {}'.format(name))


class StackNotUpdated(ManausException):
    """
    Error raised when the CloudFormation Stack is not updated because no changes
    are needed.
    """

    def __init__(self, name: str):
        super().__init__('CloudFormation Stack not updated: {}'.format(name))


class VPCError(ManausException, AttributeError):
    """
    Error raised when there are issues with VPCs configuration
    """

    def __init__(self, message: str, number_of_vpcs: int=None):
        super().__init__(message)
        self.number_of_vpcs = number_of_vpcs
